import json
from textwrap import dedent

from strif import StringTemplate, replace_multiple

from kash.config.logger import get_logger
from kash.exec import kash_action
from kash.exec.preconditions import has_html_body, has_simple_text_body
from kash.kits.media.video.speaker_labels import find_speaker_labels
from kash.llm_utils import LLM, Message, MessageTemplate
from kash.llm_utils.fuzzy_parsing import fuzzy_parse_json
from kash.llm_utils.llm_completion import llm_template_completion
from kash.media_base.timestamp_citations import html_speaker_id_span
from kash.model import Item, ItemType
from kash.utils.errors import ApiResultError, InvalidInput

log = get_logger(__name__)


@kash_action(
    precondition=has_simple_text_body | has_html_body,
)
def identify_speakers(item: Item) -> Item:
    """
    Identify speakers in a transcript and replace placeholders with their names.
    """
    if not item.body:
        raise InvalidInput("Item must have a body")

    # Find all speaker labels and their offsets
    speaker_labels = find_speaker_labels(item.body)
    if not speaker_labels:
        log.warning("This document has no speaker labels! Skipping this action.")
        return item  # No changes needed.

    # Prepare the system message and template for LLM.
    system_message = Message("You are an assistant that identifies speakers in transcripts.")
    message_template = StringTemplate(
        """
        The transcript below includes speakers identified by IDs like 'SPEAKER 0' or 'SPEAKER 1'.
        Based on the info below and the transcript, provide a mapping from speaker IDs to
        actual speaker names.

        The mapping should be in JSON format.
        If you are not sure from the content, leave the names as is and only fill in the
        known names. Examples:
        {json_examples}

        First, here is the available info on the original recording or video:

        Title: {title}
        Description: {description}

        Transcript:

        """,
        allowed_fields=["title", "description", "json_examples"],
    )

    json_examples = dedent(
        """
        Example 1: {{"0": "Alice", "1": "Bob"}}

        Example 2: {{"0": "Alice", "1": "SPEAKER 1"}}

        Example 3: {{"0": "SPEAKER 0", "1": "SPEAKER 1"}}
        """
    )

    message = message_template.format(
        title=item.title, description=item.description, json_examples=json_examples
    )

    # Perform LLM completion to get the speaker mapping.
    mapping_str = llm_template_completion(
        model=LLM.gpt_4o_mini,
        system_message=system_message,
        input=item.body,
        body_template=MessageTemplate(message + "\n\n" + "{body}"),
    ).content

    # Parse the mapping.
    try:
        speaker_mapping = fuzzy_parse_json(mapping_str)
        if not speaker_mapping:
            log.error("Could not parse speaker mapping: %s", mapping_str)
            raise ApiResultError("Could not parse speaker mapping")
        log.message("Identified speakers from transcript: %s", speaker_mapping)
    except json.JSONDecodeError as e:
        raise ApiResultError(f"Failed to parse speaker mapping from LLM output: {e}")

    # Prepare replacements.
    replacements = []
    for match in speaker_labels:
        speaker_id = match.attribute_value
        if not speaker_id:
            raise InvalidInput(f"Speaker id not found: {match}")
        new_speaker_name = speaker_mapping.get(speaker_id, f"SPEAKER {speaker_id}")
        # Prepare replacement text.
        new_span = html_speaker_id_span(f"**{new_speaker_name}:**", speaker_id)
        replacements.append((match.start_offset, match.end_offset, new_span))

    # Perform replacements.
    updated_body = replace_multiple(item.body, replacements)

    result_item = item.derived_copy(type=ItemType.doc, body=updated_body)
    return result_item



import re
from pathlib import Path
from typing import Any
from urllib.parse import parse_qs, urlparse

from frontmatter_format import to_yaml_string
from typing_extensions import override

from kash.config.logger import get_logger
from kash.config.text_styles import EMOJI_WARN
from kash.kits.media.utils.yt_dlp_tools import parse_date, ydl_download_media, ydl_extract_info
from kash.model.media_model import (
    SERVICE_YOUTUBE,
    HeatmapValue,
    MediaMetadata,
    MediaService,
    MediaUrlType,
)
from kash.utils.common.type_utils import not_none
from kash.utils.common.url import Url
from kash.utils.common.url_slice import Slice
from kash.utils.errors import ApiResultError, InvalidInput
from kash.utils.file_utils.file_formats_model import MediaType

log = get_logger(__name__)


VIDEO_ID_PATTERN = re.compile(r"^[a-zA-Z0-9_-]{11}$")


class YouTube(MediaService):
    @override
    def canonicalize_and_type(self, url: Url) -> tuple[Url | None, MediaUrlType | None]:
        parsed_url = urlparse(url)
        path_parts = parsed_url.path.strip("/").split("/")

        if parsed_url.hostname == "youtu.be":
            video_id = self.get_media_id(url)
            if video_id:
                return Url(f"https://www.youtube.com/watch?v={video_id}"), MediaUrlType.video
        elif parsed_url.hostname in ("www.youtube.com", "youtube.com", "m.youtube.com"):
            # Check for channel URLs first, as they have distinct paths
            if (
                len(path_parts) > 0
                and path_parts[0] in ("channel", "c", "user")
                or parsed_url.path.startswith("/@")
            ):
                # It's already a canonical channel URL or a recognized format.
                # TODO: Consider canonicalizing /c/ and /user/ to /@handle if possible?
                return url, MediaUrlType.channel

            query = parse_qs(parsed_url.query)

            # Check for playlist URLs
            if len(path_parts) > 0 and path_parts[0] == "playlist":
                list_id = query.get("list", [""])[0]
                if list_id:
                    return (
                        Url(f"https://www.youtube.com/playlist?list={list_id}"),
                        MediaUrlType.playlist,
                    )

            # Check for /live/ and /shorts/ paths
            if len(path_parts) == 2 and path_parts[0] in ("live", "shorts"):
                video_id = path_parts[1]
                if VIDEO_ID_PATTERN.match(video_id):
                    return Url(f"https://www.youtube.com/watch?v={video_id}"), MediaUrlType.video

            # Fallback to checking ?v= query parameter for standard video URLs
            video_id = query.get("v", [""])[0]
            if video_id and VIDEO_ID_PATTERN.match(video_id):
                return Url(f"https://www.youtube.com/watch?v={video_id}"), MediaUrlType.video

        # If none of the above matched
        return None, None

    @override
    def get_media_id(self, url: Url) -> str | None:
        parsed_url = urlparse(url)
        path_parts = parsed_url.path.strip("/").split("/")

        if parsed_url.hostname == "youtu.be":
            video_id = parsed_url.path[1:]
            if VIDEO_ID_PATTERN.match(video_id):
                return video_id
        elif parsed_url.hostname in ("www.youtube.com", "youtube.com", "m.youtube.com"):
            # Check paths first
            if len(path_parts) == 2 and path_parts[0] in ("live", "shorts"):
                video_id = path_parts[1]
                if VIDEO_ID_PATTERN.match(video_id):
                    return video_id

            # Check query parameter
            query = parse_qs(parsed_url.query)
            video_id = query.get("v", [""])[0]
            if video_id and VIDEO_ID_PATTERN.match(video_id):
                return video_id
        return None

    @override
    def thumbnail_url(self, url: Url) -> Url | None:
        id = self.get_media_id(url)
        return Url(f"https://img.youtube.com/vi/{id}/sddefault.jpg") if id else None
        # Others:
        # https://img.youtube.com/vi/{id}/hqdefault.jpg
        # https://img.youtube.com/vi/{id}/maxresdefault.jpg

    @override
    def timestamp_url(self, url: Url, timestamp: float) -> Url:
        canon_url = self.canonicalize(url)
        if not canon_url:
            raise InvalidInput(f"Unrecognized YouTube URL: {url}")
        return Url(canon_url + f"&t={timestamp}s")

    @override
    def download_media(
        self,
        url: Url,
        target_dir: Path,
        *,
        media_types: list[MediaType] | None = None,
        slice: Slice | None = None,
    ) -> dict[MediaType, Path]:
        url = not_none(self.canonicalize(url), "Not a recognized YouTube URL")
        return ydl_download_media(url, target_dir, media_types=media_types, slice=slice)

    def _extract_info(self, url: Url) -> dict[str, Any]:
        url = not_none(self.canonicalize(url), "Not a recognized YouTube URL")
        return ydl_extract_info(url)

    @override
    def metadata(self, url: Url, full: bool = False) -> MediaMetadata:
        url = not_none(self.canonicalize(url), "Not a recognized YouTube URL")
        yt_result: dict[str, Any] = self._extract_info(url)

        return self._parse_metadata(yt_result, full=full)

    @override
    def list_channel_items(self, url: Url) -> list[MediaMetadata]:
        """
        Get all video URLs and metadata from a YouTube channel or playlist.
        """

        result = self._extract_info(url)

        if "entries" in result:
            entries = result["entries"]
        else:
            log.warning("%s No videos found in the channel.", EMOJI_WARN)
            entries = []

        video_meta_list: list[MediaMetadata] = []

        # TODO: Inspect and collect rest of the metadata here, like upload date etc.
        for value in entries:
            if "entries" in value:
                # For channels there is a list of values each with their own videos.
                video_meta_list.extend(self._parse_metadata(e) for e in value["entries"])
            else:
                # For playlists, entries holds the videos.
                video_meta_list.append(self._parse_metadata(value))

        log.message("Found %d videos in channel %s", len(video_meta_list), url)

        return video_meta_list

    def _parse_metadata(
        self, yt_result: dict[str, Any], full: bool = False, **overrides: dict[str, Any]
    ) -> MediaMetadata:
        try:
            media_id = yt_result["id"]  # Renamed for clarity.
            if not media_id:
                raise KeyError("No ID found")

            url = yt_result.get("webpage_url") or yt_result.get("url")
            if not url:
                raise KeyError("No URL found")

            thumbnail_url = self.thumbnail_url(Url(url))
            # thumbnail_url = best_thumbnail(yt_result)  # Alternate approach, but messier.

            # FIXME: upload_date not working?
            # Apparently upload_date is in full video metadata but not channel metadata.
            upload_date_str = yt_result.get("upload_date")
            upload_date = parse_date(upload_date_str) if upload_date_str else None

            # Heatmap is interesting but verbose so skipping by default.
            heatmap = None
            if full:
                heatmap = [HeatmapValue(**h) for h in yt_result.get("heatmap", [])] or None

            result = MediaMetadata(
                media_id=media_id,
                media_service=SERVICE_YOUTUBE,
                url=url,
                thumbnail_url=thumbnail_url,
                title=yt_result["title"],
                description=yt_result["description"],
                upload_date=upload_date,
                channel_url=Url(yt_result["channel_url"]),
                view_count=yt_result.get("view_count"),
                duration=yt_result.get("duration"),
                heatmap=heatmap,
                **overrides,
            )
            log.message("Parsed YouTube metadata: %s", result)
        except KeyError as e:
            log.error("Missing key in YouTube metadata (see saved object): %s", e)
            log.save_object(
                "yt_dlp result", None, to_yaml_string(yt_result, stringify_unknown=True)
            )
            raise ApiResultError("Did not find key in YouTube metadata: %s" % e)

        return result


def best_thumbnail(data: dict[str, Any]) -> Url | None:
    """
    Get the best thumbnail from YouTube metadata, which is of the form:
    {
        'thumbnails': [
            {'url': 'https://i.ytimg.com/vi/gc417NquXbk/hqdefault.jpg?sqp=-oaymwEbCKgBEF5IVfKriqkDDggBFQAAiEIYAXABwAEG&rs=AOn4CLC7F70CUSkwqkrgEwKX1AmXCJ8jsQ', 'height': 94, 'width': 168},
            {'url': 'https://i.ytimg.com/vi/gc417NquXbk/hqdefault.jpg?sqp=-oaymwEbCMQBEG5IVfKriqkDDggBFQAAiEIYAXABwAEG&rs=AOn4CLA6pIOQRUlixQogTAR0NAv3zJgxqQ', 'height': 110, 'width': 196},
            {'url': 'https://i.ytimg.com/vi/gc417NquXbk/hqdefault.jpg?sqp=-oaymwEcCPYBEIoBSFXyq4qpAw4IARUAAIhCGAFwAcABBg==&rs=AOn4CLC7zW8Hu59MgQkRPX4WsVpc-tkxxQ', 'height': 138, 'width': 246},
            {'url': 'https://i.ytimg.com/vi/gc417NquXbk/hqdefault.jpg?sqp=-oaymwEcCNACELwBSFXyq4qpAw4IARUAAIhCGAFwAcABBg==&rs=AOn4CLB7nBpvNKwCwrtr_lv85T0GITOFjA', 'height': 188, 'width': 336},
        ],
    }
    """
    thumbnail_url = None
    try:
        thumbnails = data["thumbnails"]
        if not isinstance(thumbnails, list):
            return None
        largest_thumbnail = max(thumbnails, key=lambda x: x.get("width", 0))
        thumbnail_url = largest_thumbnail.get("url", None)
    except (KeyError, TypeError):
        pass

    if not thumbnail_url:
        thumbnail_url = data.get("thumbnail")

    return Url(thumbnail_url) if thumbnail_url else None


## Tests


def test_canonicalize_youtube():
    youtube = YouTube()

    def assert_canon(url: str, expected_canon_url: str | None, expected_type: MediaUrlType | None):
        canon_url, url_type = youtube.canonicalize_and_type(Url(url))
        assert canon_url == (Url(expected_canon_url) if expected_canon_url else None)
        assert url_type == expected_type
        # Also test canonicalize convenience method
        if expected_canon_url:
            assert youtube.canonicalize(Url(url)) == Url(expected_canon_url)
        else:
            assert youtube.canonicalize(Url(url)) is None

    # Video URLs
    assert_canon(
        "https://youtu.be/dQw4w9WgXcQ",
        "https://www.youtube.com/watch?v=dQw4w9WgXcQ",
        MediaUrlType.video,
    )
    assert_canon(
        "https://www.youtube.com/watch?v=dQw4w9WgXcQ",
        "https://www.youtube.com/watch?v=dQw4w9WgXcQ",
        MediaUrlType.video,
    )
    assert_canon(
        "https://m.youtube.com/watch?v=dQw4w9WgXcQ&feature=share",
        "https://www.youtube.com/watch?v=dQw4w9WgXcQ",
        MediaUrlType.video,
    )
    assert_canon(
        "https://www.youtube.com/watch?v=_5y0AalUDh4&list=PL9XbNw3iJu1zKJRyV3Jz3rqlFV1XJfvNv&index=12",
        "https://www.youtube.com/watch?v=_5y0AalUDh4",
        MediaUrlType.video,
    )

    # Live URLs
    assert_canon(
        "https://www.youtube.com/live/XVwpL_cAvrw?si=mgv-xnhaO6UJLxwb",
        "https://www.youtube.com/watch?v=XVwpL_cAvrw",
        MediaUrlType.video,
    )

    # Shorts URLs
    assert_canon(
        "https://www.youtube.com/shorts/abcdefghijk",  # Example valid ID format
        "https://www.youtube.com/watch?v=abcdefghijk",
        MediaUrlType.video,
    )
    assert_canon(
        "https://youtube.com/shorts/lmnopqrs_tu",  # Example valid ID format (11 chars)
        "https://www.youtube.com/watch?v=lmnopqrs_tu",
        MediaUrlType.video,
    )

    # Channel URLs
    assert_canon(
        "https://www.youtube.com/@hubermanlab",
        "https://www.youtube.com/@hubermanlab",
        MediaUrlType.channel,
    )
    assert_canon(
        "https://www.youtube.com/c/inanutshell",
        "https://www.youtube.com/c/inanutshell",
        MediaUrlType.channel,
    )
    assert_canon(
        "https://www.youtube.com/channel/UCXuqSBlHAE6Xw-yeJA0Tunw",
        "https://www.youtube.com/channel/UCXuqSBlHAE6Xw-yeJA0Tunw",
        MediaUrlType.channel,
    )
    assert_canon(
        "https://www.youtube.com/user/Vsauce",
        "https://www.youtube.com/user/Vsauce",
        MediaUrlType.channel,
    )

    # Playlist URLs
    assert_canon(
        "https://youtube.com/playlist?list=PLPNW_gerXa4N_PVVoq0Za03YKASSGCazr&si=9IVO8p-ZwmMLI18F",
        "https://www.youtube.com/playlist?list=PLPNW_gerXa4N_PVVoq0Za03YKASSGCazr",
        MediaUrlType.playlist,
    )
    assert_canon(
        "https://www.youtube.com/playlist?list=PLFsQleAWXsj_4yDeebiIADdH5FMayBiJo",
        "https://www.youtube.com/playlist?list=PLFsQleAWXsj_4yDeebiIADdH5FMayBiJo",
        MediaUrlType.playlist,
    )

    # Invalid/Unrecognized URLs
    assert_canon("https://example.com", None, None)
    assert_canon("https://www.youtube.com/", None, None)
    assert_canon(
        "https://www.youtube.com/feed/subscriptions", None, None
    )  # Example non-content URL
    assert_canon("https://youtu.be/", None, None)  # Missing ID
    assert_canon("https://www.youtube.com/watch?list=abc", None, None)  # Missing v= parameter



from chopdiff.divs import parse_divs

import kash.kits.docs.doc_formats  # noqa: F401  # Ensure all media tools are available.
from kash.config.logger import get_logger
from kash.exec import kash_action
from kash.exec.preconditions import is_audio_resource, is_url_resource, is_video_resource
from kash.media_base.media_tools import cache_and_transcribe
from kash.model import FileExt, Format, Item, ItemType, common_params
from kash.utils.common.type_utils import not_none
from kash.utils.common.url import as_file_url
from kash.workspaces import current_ws

log = get_logger(__name__)


@kash_action(
    precondition=is_url_resource | is_audio_resource | is_video_resource,
    params=common_params("language"),
    mcp_tool=True,
)
def transcribe(item: Item, language: str = "en") -> Item:
    """
    Download and transcribe audio from a podcast or video and return raw text,
    including timestamps if available (as HTML `<span>` tags), also caching
    video, audio, and transcript as local files.
    """

    if item.url:
        url = item.url
    else:
        url = as_file_url(current_ws().base_dir / not_none(item.store_path))

    transcription = cache_and_transcribe(url, language=language)

    result_item = item.derived_copy(
        type=ItemType.doc,
        body=transcription,
        format=Format.html,  # Important to note this since we put in timestamp span tags.
        file_ext=FileExt.html,
        external_path=None,
    )

    log.message("Got transcription: %s", parse_divs(transcription).size_summary())

    return result_item



from pathlib import Path

from kash.exec import import_and_register

import_and_register(
    __package__,
    Path(__file__).parent,
    [
        "actions",
        "media_services",
        "video",
    ],
)



from textwrap import indent

from chopdiff.docs import (
    BOF_TOK,
    EOF_TOK,
    PARA_BR_TOK,
    SENT_BR_TOK,
    SentIndex,
    TextDoc,
    TextUnit,
    TokenMapping,
    search_tokens,
)
from chopdiff.html import ContentNotFound, TimestampExtractor

from kash.config.logger import get_logger
from kash.exec import kash_action
from kash.exec.preconditions import has_simple_text_body, has_timestamps
from kash.media_base.timestamp_citations import add_citation_to_text, format_timestamp_citation
from kash.model import Format, Item, ItemType, common_params
from kash.utils.common.format_utils import fmt_loc
from kash.utils.common.type_utils import not_none
from kash.utils.errors import InvalidInput, UnexpectedError
from kash.workspaces.source_items import find_upstream_item, find_upstream_resource

log = get_logger(__name__)


@kash_action(
    precondition=has_simple_text_body & ~has_timestamps,
    params=common_params("chunk_unit"),
)
def backfill_timestamps(item: Item, chunk_unit: TextUnit = TextUnit.paragraphs) -> Item:
    """
    Backfill timestamps from a source document.
    Seeks through the document this doc is derived from for timestamps and inserts them
    into the text of the current doc. Source must have similar tokens.
    """
    if chunk_unit not in (TextUnit.sentences, TextUnit.paragraphs):
        raise InvalidInput(f"Only support sentences and paragraphs for chunk unit: {chunk_unit}")

    if chunk_unit == TextUnit.sentences:
        citation_tokens = [SENT_BR_TOK, PARA_BR_TOK, EOF_TOK]
    elif chunk_unit == TextUnit.paragraphs:
        citation_tokens = [PARA_BR_TOK, EOF_TOK]
    else:
        raise UnexpectedError(f"Invalid text unit: {chunk_unit}")

    source_item = find_upstream_item(item, has_timestamps)
    # Find the original resource (video or audio) this timestamped item came from.
    orig_resource = find_upstream_resource(source_item)
    source_path = orig_resource.store_path
    source_url = orig_resource.url

    if not item.body:
        raise InvalidInput(f"Item must have a body: {item}")
    if not source_item.body:
        raise InvalidInput(f"Source item must have a body: {source_item}")
    if not source_path:
        raise InvalidInput(f"Source item must have a store path: {source_item}")
    if not source_url:
        log.warning(
            "Source item has no URL, so will not create timestamp hotlinks: %s", source_item
        )

    log.message(
        "Pulling timestamps from source item: %s", fmt_loc(not_none(source_item.store_path))
    )

    # Parse current doc.
    item_doc = TextDoc.from_text(item.body)
    item_wordtoks = list(item_doc.as_wordtoks(bof_eof=True))

    # Don't bother parsing sentences on the source document, which may be long and with HTML.
    extractor = TimestampExtractor(source_item.body)
    source_wordtoks = extractor.wordtoks

    log.message(
        "Mapping source doc with %s wordtoks back to this item, %s.",
        len(source_wordtoks),
        item_doc.size_summary(),
    )

    token_mapping = TokenMapping(source_wordtoks, item_wordtoks)

    log.info(
        "Timestamp extractor mapping diff:\n%s",
        indent(token_mapping.diff.as_diff_str(include_equal=False), prefix="    "),
    )

    log.save_object("Token mapping", None, token_mapping.full_mapping_str())

    output_item = item.derived_copy(type=ItemType.doc, format=Format.md_html)

    sent_index_list: list[SentIndex] = []
    timestamp_list: list[float] = []

    for wordtok_offset, (wordtok, sent_index) in enumerate(
        item_doc.as_wordtok_to_sent(bof_eof=True)
    ):
        if wordtok in citation_tokens:
            # If we're inserting citations at paragraph breaks, we need to back up to the beginning of the paragraph.
            # If we're inserting citations at sentence breaks, we can just use the per-sentence timestamps.
            if chunk_unit == TextUnit.paragraphs:
                start_para_index, start_para_wordtok = (
                    search_tokens(item_wordtoks)
                    .at(wordtok_offset)
                    .seek_back([BOF_TOK, PARA_BR_TOK])
                    .next()
                    .get_token()
                )

                log.info(
                    "Searching to previous para break behind %s (%s) got %s (%s)",
                    wordtok_offset,
                    wordtok,
                    start_para_index,
                    start_para_wordtok,
                )

                wordtok_offset = start_para_index

            source_wordtok_offset = token_mapping.map_back(wordtok_offset)

            log.info(
                "Mapping token at offset back to source doc: %s (%s) -> %s (%s)",
                wordtok_offset,
                wordtok,
                source_wordtok_offset,
                source_wordtoks[source_wordtok_offset],
            )

            try:
                timestamp, _index, _offset = extractor.extract_preceding(source_wordtok_offset)

                timestamp_list.append(timestamp)
                sent_index_list.append(sent_index)

                sent = item_doc.get_sent(sent_index)
                if sent.is_markup():
                    log.info("Skipping markup-only sentence: %s", sent.text)
                    continue

                sent.text = add_citation_to_text(
                    sent.text,
                    format_timestamp_citation(source_url, source_path, timestamp),
                )
            except ContentNotFound:
                # Missing timestamps aren't fatal since it might be meta text like "Speaker 1:".
                log.warning(
                    "Failed to extract timestamp at doc token %s (%s) -> source token %s (%s): %s",
                    wordtok_offset,
                    wordtok,
                    source_wordtok_offset,
                    source_wordtoks[source_wordtok_offset],
                    sent_index,
                )

    first = timestamp_list[0] if timestamp_list else "none"
    last = timestamp_list[-1] if timestamp_list else "none"
    log.message(
        "Found %s timestamps in source doc from %s to %s.", len(timestamp_list), first, last
    )

    output_item.body = item_doc.reassemble()

    return output_item



from chopdiff.divs import parse_divs
from chopdiff.docs import search_tokens
from chopdiff.html import TimestampExtractor, html_img, md_para
from strif import Insertion, insert_multiple

from kash.config.logger import get_logger
from kash.exec import kash_action, kash_precondition
from kash.exec.preconditions import has_simple_text_body, has_timestamps
from kash.kits.media.video.image_similarity import filter_similar_frames
from kash.kits.media.video.video_frames import capture_frames
from kash.model import Format, Item, ItemType, Param
from kash.utils.common.format_utils import fmt_loc
from kash.utils.common.url import as_file_url
from kash.utils.errors import ContentError, InvalidInput
from kash.utils.file_utils.file_formats_model import MediaType
from kash.web_content.file_cache_utils import cache_file, cache_resource
from kash.workspaces import current_ws
from kash.workspaces.source_items import find_upstream_resource

log = get_logger(__name__)


FRAME_CAPTURE = "frame-capture"
"""Class name for a frame capture from a video."""


@kash_precondition
def has_frame_captures(item: Item) -> bool:
    return bool(item.body and item.body.find(f'<img class="{FRAME_CAPTURE}">') != -1)


@kash_action(
    precondition=has_simple_text_body & has_timestamps & ~has_frame_captures,
    params=(
        Param(
            "threshold",
            "The similarity threshold for filtering consecutive frames.",
            type=float,
            default_value=0.6,
        ),
    ),
)
def insert_frame_captures(item: Item, threshold: float = 0.6) -> Item:
    """
    Look for timestamped video links and insert frame captures after each one.
    """
    if not item.body:
        raise InvalidInput("Item has no body")

    # Find the original video resource.
    orig_resource = find_upstream_resource(item)
    paths = cache_resource(orig_resource)
    if MediaType.video not in paths:
        raise InvalidInput(f"Item has no video: {item}")
    video_path = paths[MediaType.video]

    # Extract all timestamps.
    extractor = TimestampExtractor(item.body)
    timestamp_matches = list(extractor.extract_all())

    log.message(
        f"Found {len(timestamp_matches)} timestamps in the document, {parse_divs(item.body).size_summary()}."
    )

    # Extract frame captures, and put them in the workspace's assets directory.
    target_dir = current_ws().assets_dir
    timestamps = [timestamp for timestamp, _index, _offset in timestamp_matches]
    frame_paths = capture_frames(video_path, timestamps, target_dir, prefix=item.slug_name())

    # Save images in file cache for later as well.
    for frame_path in frame_paths:
        cache_file(frame_path)
    log.message(f"Saved {len(frame_paths)} frame captures to cache.")

    # Filter out similar consecutive frames.
    unique_indices = filter_similar_frames(frame_paths, threshold)
    unique_frame_paths = [frame_paths[i] for i in unique_indices]
    unique_matches = [timestamp_matches[i] for i in unique_indices]

    log.message(
        f"Filtered out {len(frame_paths) - len(unique_frame_paths)}/{len(frame_paths)} similar frames."
    )
    log.message(
        f"Extracted {len(unique_frame_paths)} unique frame captures to: {fmt_loc(target_dir)}"
    )

    # Create a set of indices that were kept.
    kept_indices = set(unique_indices)

    # Prepare insertions.
    log.message(
        "Inserting %s frame captures, have %s wordtoks",
        len(unique_matches),
        len(extractor.offsets),
    )
    insertions: list[Insertion] = []
    for i, (timestamp, index, offset) in enumerate(timestamp_matches):
        # Only process timestamps whose frames weren't filtered out
        if i not in kept_indices:
            continue

        try:
            insert_index = (
                search_tokens(extractor.wordtoks)
                .at(index)
                .seek_forward(["</span>"])
                .next()
                .get_index()
            )
        except KeyError:
            raise ContentError(
                f"No matching tag close starting at {offset}: {extractor.wordtoks[offset:]}"
            )

        new_offset = extractor.offsets[insert_index]
        frame_path = frame_paths[i]
        insertions.append(
            (
                new_offset,
                md_para(
                    html_img(
                        as_file_url(frame_path),  # TODO: Serve these.
                        f"Frame at {timestamp} seconds",
                        class_name=FRAME_CAPTURE,
                    )
                ),
            )
        )

    # Insert img tags into the document.
    output_text = insert_multiple(item.body, insertions)

    # Create output item.
    output_item = item.derived_copy(type=ItemType.doc, format=Format.md_html)
    output_item.body = output_text

    return output_item



import re
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from typing_extensions import override

from kash.config.logger import get_logger
from kash.kits.media.utils.yt_dlp_tools import parse_date, ydl_download_media, ydl_extract_info
from kash.model.media_model import MediaMetadata, MediaService, MediaUrlType, Slice
from kash.utils.common.type_utils import not_none
from kash.utils.common.url import Url
from kash.utils.errors import ApiResultError, InvalidInput
from kash.utils.file_utils.file_formats_model import MediaType

log = get_logger(__name__)


VIDEO_PATTERN = r"^/(\d+)$"
CHANNEL_PATTERN = r"^/([a-zA-Z0-9_-]+)$"


class Vimeo(MediaService):
    @override
    def canonicalize_and_type(self, url: Url) -> tuple[Url | None, MediaUrlType | None]:
        parsed_url = urlparse(url)
        if parsed_url.hostname == "vimeo.com":
            path = parsed_url.path
            video_match = re.match(VIDEO_PATTERN, path)
            if video_match:
                return Url(f"https://vimeo.com/{video_match.group(1)}"), MediaUrlType.video
            channel_match = re.match(CHANNEL_PATTERN, path)
            if channel_match:
                return Url(f"https://vimeo.com/{channel_match.group(1)}"), MediaUrlType.channel
        return None, None

    @override
    def get_media_id(self, url: Url) -> str | None:
        parsed_url = urlparse(url)
        if parsed_url.hostname == "vimeo.com":
            path = parsed_url.path
            video_match = re.match(VIDEO_PATTERN, path)
            if video_match:
                return f"video:{video_match.group(1)}"
            channel_match = re.match(CHANNEL_PATTERN, path)
            if channel_match:
                return f"channel:{channel_match.group(1)}"
        return None

    @override
    def metadata(self, url: Url, full: bool = False) -> MediaMetadata:
        url = not_none(self.canonicalize(url), "Not a recognized Vimeo URL")
        vimeo_result: dict[str, Any] = self._extract_info(url)
        return self._parse_metadata(vimeo_result, full=full)

    @override
    def thumbnail_url(self, url: Url) -> Url | None:
        vimeo_result = self._extract_info(url)
        thumbnails = vimeo_result.get("thumbnails", [])
        if thumbnails:
            return Url(thumbnails[-1]["url"])  # Get the last (usually highest quality) thumbnail
        return None

    @override
    def timestamp_url(self, url: Url, timestamp: float) -> Url:
        canon_url, url_type = self.canonicalize_and_type(url)
        if not canon_url:
            raise InvalidInput(f"Unrecognized Vimeo URL: {url}")
        if url_type == MediaUrlType.video:
            return Url(f"{canon_url}#t={timestamp}")
        return canon_url  # For channels, just return the canonical URL

    @override
    def download_media(
        self,
        url: Url,
        target_dir: Path,
        *,
        media_types: list[MediaType] | None = None,
        slice: Slice | None = None,
    ) -> dict[MediaType, Path]:
        url = not_none(self.canonicalize(url), "Not a recognized Vimeo URL")
        return ydl_download_media(url, target_dir, media_types=media_types, slice=slice)

    @override
    def list_channel_items(self, url: Url) -> list[MediaMetadata]:
        raise NotImplementedError()

    def _extract_info(self, url: Url) -> dict[str, Any]:
        url = not_none(self.canonicalize(url), "Not a recognized Vimeo URL")
        return ydl_extract_info(url)

    def _parse_metadata(
        self, vimeo_result: dict[str, Any], full: bool = False, **overrides: dict[str, Any]
    ) -> MediaMetadata:
        try:
            media_id = vimeo_result["id"]
            if not media_id:
                raise KeyError("No ID found")

            url = vimeo_result.get("webpage_url") or vimeo_result.get("url")
            if not url:
                raise KeyError("No URL found")

            thumbnail_url = self.thumbnail_url(Url(url))

            upload_date_str = vimeo_result.get("upload_date")
            upload_date = parse_date(upload_date_str) if upload_date_str else None

            _, url_type = self.canonicalize_and_type(Url(url))

            result = MediaMetadata(
                media_id=media_id,
                media_service="vimeo",
                url=url,
                thumbnail_url=thumbnail_url,
                title=vimeo_result["title"],
                description=vimeo_result.get("description"),
                upload_date=upload_date,
                channel_url=Url(vimeo_result.get("uploader_url", "")),
                view_count=vimeo_result.get("view_count"),
                duration=vimeo_result.get("duration") if url_type == MediaUrlType.video else None,
                heatmap=None,
                **overrides,
            )
            log.message("Parsed Vimeo metadata: %s", result)
        except KeyError as e:
            log.error("Missing key in Vimeo metadata: %s", e)
            raise ApiResultError(f"Did not find key in Vimeo metadata: {e}")

        return result
